import pika
import uuid
import json


class RPCMethod(object):
    def __init__(self, service):
        self.service = service
        self.corr_id = None
        self.response = None
        self.connection = pika.BlockingConnection(
            pika.ConnectionParameters(host="localhost")
        )
        self.channel = self.connection.channel()
        result = self.channel.queue_declare(queue="", exclusive=True)
        self.callback_queue = result.method.queue

        self.channel.basic_consume(
            queue=self.callback_queue,
            on_message_callback=self.on_response,
            auto_ack=True,
        )

    def on_response(self, ch, method, props, body):
        if self.corr_id == props.correlation_id:
            self.response = body

    def call(self, method, *args):
        body = {"method": method, "args": args}

        self.response = None
        self.corr_id = str(uuid.uuid4())
        self.channel.basic_publish(
            exchange="",
            routing_key=self.service,
            properties=pika.BasicProperties(
                reply_to=self.callback_queue, correlation_id=self.corr_id
            ),
            body=json.dumps(body),
        )

        while self.response is None:
            self.connection.process_data_events()

        return json.loads(self.response)["result"]


class Service:
    def __init__(self, service_name):
        self.service_name = service_name
        self.connection = pika.BlockingConnection(
            pika.ConnectionParameters("localhost")
        )
        self.channel = self.connection.channel()
        self.methods = {}
        self.setup_queues()

    def setup_queues(self):
        self.channel.queue_declare(queue=self.service_name)

    def method(self):
        def wrapper(fn):
            self.methods[fn.__qualname__] = fn
            return None

        return wrapper

    def _callback(self, ch, method, props, body):
        payload = json.loads(body)
        if payload["method"] in self.methods:
            result = self.methods[payload["method"]](*payload["args"])
            body = {"type": "response", "result": result}
            ch.basic_publish(
                exchange="",
                routing_key=props.reply_to,
                properties=pika.BasicProperties(correlation_id=props.correlation_id),
                body=json.dumps(body),
            )

    def start(self):
        self.channel.basic_consume(
            queue=self.service_name, auto_ack=True, on_message_callback=self._callback
        )

        self.channel.start_consuming()
